from typing import Optional, Any, Union
from venv import create
from fastapi import FastAPI, status, HTTPException
import uvicorn
from sqlalchemy import inspect
from contextlib import contextmanager
from pprint import pprint

from biosearch.FastAPI.models import Base, Person, Presentation, Slide, Textbox
from biosearch.FastAPI.schemas import PersonSerializer, PresentationSerializer, SlideSerializer, TextboxSerializer
from biosearch.FastAPI.database import SessionLocal, engine

Base.metadata.create_all(bind=engine)

app = FastAPI()

@contextmanager
def create_db():
    """ Provide transactional scope around a series of operations """
    db = SessionLocal()
    try:
        yield db
        db.commit()
    except:
        db.rollback()
        raise
    finally:
        db.close()

@app.get('/')# ?name=world
def example(name : Optional[str] = 'default'):
    return {'Optional name':f'Hello {name}'}

@app.get('/persons', response_model = list[PersonSerializer], status_code = status.HTTP_200_OK)
def retrieve_persons():
    with create_db() as db:
        people = db.query(Person).all()
    # print(inspect(people).expired)
    # pprint(people)
    return people


@app.get('/person/{id}', response_model = PersonSerializer, status_code = status.HTTP_200_OK)
def retrieve_person_by_id(id : int):
    with create_db() as db:
        retrieved_person =db.query(Person).filter(Person.id == id).first()
    
    if retrieved_person == None:
        raise HTTPException(status_code = status.HTTP_404_NOT_FOUND, detail = 'Resource not found')

    return retrieved_person 

@app.post('/person', response_model = PersonSerializer, status_code = status.HTTP_201_CREATED)
def create_new_person(person:PersonSerializer):
    with create_db() as db:
        new_person = Person(
            first_name = person.first_name,
            last_name = person.last_name,
            age = person.age,
            gender = person.gender,
            nationality = person.nationality,
            hasWatch = person.hasWatch
        )
        db.add(new_person)

    return new_person

@app.put('/person/{id}', response_model = PersonSerializer, status_code = status.HTTP_202_ACCEPTED)
def update_person(id: int, person: PersonSerializer):
    with create_db() as db:
        retrieved_person = db.query(Person).filter(Person.id == id).first()
        print(retrieved_person)
        list_attributes = ['first_name', 'last_name', 'age', 'gender', 'nationality', 'hasWatch']
        for attr in list_attributes:
            new_value = getattr(person, attr)
            setattr(retrieved_person, attr, new_value)
        
    return retrieved_person

@app.delete('/person/{id}')
def delete_person(id:int):
    with create_db() as db:
        retrieved_person = db.query(Person).filter(Person.id == id).first()
        if retrieved_person == None:
            raise HTTPException(status_code = status.HTTP_404_NOT_FOUND, detail = 'Requested person could not be deleted')
        db.delete(retrieved_person)
    return retrieved_person


@app.get("/google-presentations", response_model=list[PresentationSerializer])
def get_presentations():
    with create_db() as db:
        presentations = db.query(Presentation).all()
    return presentations

@app.get('/google-slides/{presentation_id}', response_model = list[Union[SlideSerializer, TextboxSerializer]])
def get_slides(presentation_id : int, slide_id : Union[int, None] = None): # ? add the ability to enter skills
    with create_db() as db: # ! extension
        if slide_id == None:
            slides = db.query(Slide).filter(Slide.presentation_id == presentation_id).all()

        else: # isinstance(slide_id, int) == True
            print(slide_id)
            slides = db.query(Textbox).filter(Textbox.slide_id == slide_id).all()

    return slides

if __name__ == '__main__':
    uvicorn.run('main:app', reload = True)


import psycopg2
import psycopg2.extras

try:
    from DB.db_config import connection_params
    import DB.db_queries as query
except Exception as e:
    print('the file is run directly and not called by other file')
    from db_config import connection_params
    import db_queries as query


def get_all_google_presentations():
    try:
        with psycopg2.connect(**connection_params) as connection:
            with connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cursor:
                cursor.execute(query.get_all_google_presentations())
                return cursor.fetchall()
    except (Exception, psycopg2.DatabaseError) as error:
        print("Failed to initialise tables due to :" + error)

def upload_file_to_temp_table(file_path):
    try:
        with psycopg2.connect(**connection_params) as connection:
            with connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cursor:
                cursor.execute(query.create_temp_table())
                cursor.execute(query.upload_data_from_file(file_path))
                cursor.execute(query.merge_temp_table_with_google_slides_table())
                cursor.execute(query.merge_temp_table_with_google_slide_textboxes_table())
                print('Merged successfully!')
    except (Exception, psycopg2.DatabaseError) as error:
        print("Failed to initialise tables due to :" + error)

def update_google_presentation_revision_id(revision_id):
    try:
        with psycopg2.connect(**connection_params) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query.update_google_presentation_revision_id(revision_id))
        print('''Record (%s) has been updated successfully into google_presentations table ''' % revision_id)
    except (Exception, psycopg2.DatabaseError) as error:
        print("Failed to update the revision_id of google_presentation table due to :" + error)

def insert_google_presentation(presentation_link, club_name, revision_id):
    record_to_insert = (presentation_link, club_name, revision_id)
    try:
        with psycopg2.connect(**connection_params) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query.insert_google_presentation(), record_to_insert)
        print('''Record (%s, %s, %s) inserted successfully into google_presentations table ''' % record_to_insert)
    except (Exception, psycopg2.DatabaseError) as error:
        print("Failed to insert to google_presentation table due to :" + error)


def initialise_database_tables():
    try:
        with psycopg2.connect(**connection_params) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query.drop_all_tables())
                cursor.execute(query.create_all_tables())
        print("Initialised all the tables successfully!")
    except (Exception, psycopg2.DatabaseError) as error:
        print("Failed to initialise tables due to :" + error)


if __name__ == "__main__":
    initialise_database_tables()
    insert_google_presentation('19Fh8FULu6hNFVpb5iK6-NrBNqin3SLozE0emdJtstPM', 'Dekker', 'Miw8XHfMw58bkw')
    get_all_google_presentations()



def drop_all_tables():
    return '''DROP TABLE IF EXISTS google_presentations CASCADE; 
              DROP TABLE IF EXISTS google_slides CASCADE; 
              DROP TABLE IF EXISTS google_slide_textboxes CASCADE;'''

def get_all_google_presentations():
    return '''SELECT presentation_id, presentation_link, revision_id FROM google_presentations;'''

def create_all_tables():
    return create_table_google_presentations() + create_table_google_slides() + create_table_google_slide_textboxes()


def create_table_google_presentations():
    return '''CREATE TABLE IF NOT EXISTS google_presentations ( 
              presentation_id SERIAL PRIMARY KEY,
	          presentation_link VARCHAR UNIQUE NOT NULL,
	          club_name VARCHAR UNIQUE NOT NULL,
	          revision_id VARCHAR NOT NULL,
	          created_at TIMESTAMP NOT NULL DEFAULT NOW(),
	          updated_at TIMESTAMP NOT NULL DEFAULT NOW()
              );'''

def update_google_presentation_revision_id(revision_id):
    return "UPDATE google_presentations SET revision_id = '" + revision_id + "'"


def insert_google_presentation():
    return '''INSERT INTO google_presentations (presentation_link, club_name, revision_id) VALUES (%s, %s, %s)'''

def create_table_google_slides():
    return '''CREATE TABLE IF NOT EXISTS google_slides ( 
                slide_id SERIAL PRIMARY KEY, 
                slide_link VARCHAR UNIQUE NOT NULL,
                presentation_id INT NOT NULL,
                created_at TIMESTAMP NOT NULL DEFAULT NOW(),
                updated_at TIMESTAMP NOT NULL DEFAULT NOW(),
                CONSTRAINT fk_presentation FOREIGN KEY(presentation_id) REFERENCES google_presentations(presentation_id) ON DELETE CASCADE
            );'''

def create_temp_table():
    return ''' CREATE TEMPORARY TABLE IF NOT EXISTS temp_t AS 
                SELECT gs.presentation_id, gs.slide_link, textbox_link, text FROM 
                google_presentations AS gp INNER JOIN 
                google_slides AS gs ON gp.presentation_id = gs.presentation_id INNER JOIN
                google_slide_textboxes AS gt ON gs.slide_id = gt.slide_id WITH NO DATA;
            '''

def upload_data_from_file(file_path):
    return "COPY temp_t FROM '" +file_path+"' (DELIMITER(';'))"

def merge_temp_table_with_google_slides_table():
    return '''INSERT INTO google_slides(slide_link, presentation_id)
              SELECT slide_link, presentation_id FROM temp_t GROUP BY slide_link, presentation_id
              ON CONFLICT DO NOTHING;'''

def merge_temp_table_with_google_slide_textboxes_table():
    return '''WITH join_t AS(
                            SELECT gs.slide_id, tm.*
                            FROM google_slides as gs
                            INNER JOIN temp_t as tm
                            ON gs.slide_link = tm.slide_link)
              INSERT INTO google_slide_textboxes(slide_id, textbox_link, text)
              SELECT slide_id, textbox_link, text FROM join_t
              ON CONFLICT (textbox_link)
              DO 
                UPDATE SET text = EXCLUDED.text;
    '''
def create_table_google_slide_textboxes():
    return '''CREATE TABLE IF NOT EXISTS google_slide_textboxes ( 
                textbox_id SERIAL PRIMARY KEY, 
                textbox_link VARCHAR UNIQUE NOT NULL, 
                slide_id INT NOT NULL, 
                text VARCHAR NOT NULL, 
                created_at TIMESTAMP NOT NULL DEFAULT NOW(), 
                updated_at TIMESTAMP NOT NULL DEFAULT NOW(), 
                CONSTRAINT fk_slide_id FOREIGN KEY(slide_id) REFERENCES google_slides(slide_id) ON DELETE CASCADE
            );
            ALTER TABLE google_slide_textboxes ADD COLUMN ts tsvector
            GENERATED ALWAYS AS (to_tsvector('english', text)) STORED;
	        CREATE INDEX ts_idx ON google_slide_textboxes USING GIN (ts);'''



import os
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

USER = os.environ.get('POSTGRESQL_USERNAME')
POSTGRESQL_PASSWORD = os.environ.get('PASSWORD', default='')
DOMAIN = 'localhost'
DATABASE = 'postgres'

SQLALCHEMY_DATABASE_URL = f"postgresql://{USER}:{POSTGRESQL_PASSWORD}@{DOMAIN}/{DATABASE}"
engine = create_engine(SQLALCHEMY_DATABASE_URL)
SessionLocal = sessionmaker(autocommit=False, expire_on_commit = False, autoflush=False, bind=engine) # creates the factory